//-----------------------------LICENSE NOTICE------------------------------------
//  This file is part of Laser Boy
//  Copyright (C) 2017 Carlos Sevila (@CarlosSevila)
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//------------------------------------------------------------------------------

#include "game.h"

u8 map[MAP_WIDTH][MAP_HEIGHT];
u8 initialMap;
u8 currentMap;
TPlayer player;
TMovingEntity enemies[MAX_ENEMIES];
TEntity deaths[MAX_ENEMIES];
TMovingEntity playerBullets[PLAYER_MAX_BULLETS];
TBlock blocks[MAX_BLOCKS];
u8 numBlocks;
u8 timeCount;
u8 seconds;
u16 minutes;
u8 deadCount;
u8 winCount;
u16 totalDefeats;
u8 isPaused;
u8 pauseTicks;
u8 drawPause;

void InitializeGame(void) {
	deadCount = 0;
	winCount = 0;
	totalDefeats = 0;
    seconds = 0;
	minutes = 0;
	timeCount = 50;
	isPaused = 0;
	pauseTicks = 0;
	drawPause = 0;

    LoadMap(currentMap);
}

void GameLoop(void) {
    u8 ticks = 0, i, j;

    while (1) {

    	// Update
    	if (isPaused) {
    		CheckPause();

    		// The player can quit the game while paused.
			if (cpct_isKeyPressed(keys.exit)) {
				break;
			}

			// To prevent drawing.
			continue;	
    	}

    	if (!CheckDeath() && !CheckWin()) {
			CheckInput(&ticks);
	        UpdatePlayer(&ticks);
	        UpdateBlocks();

	        for (i = 0; i < PLAYER_MAX_BULLETS; i++) {
	            if (playerBullets[i].active) {
	                UpdateMovingEntity(&ticks, &playerBullets[i]);

	                // Check if the bullet hits an enemy.
	                for (j = 0; j < MAX_ENEMIES; j++) {
	                	if (enemies[j].active) {
	                		if (CheckEntitiesCollisionBB(&playerBullets[i].entity, &enemies[j].entity)) {

	                			if (!(enemies[j].type & ENEMY_TYPE_LASER_INMUNE)) {
							        enemies[j].erase = 2;
							        enemies[j].active = 0;
	                			}

						        playerBullets[i].erase = 2;
						        playerBullets[i].active = 0;
							}
	                	}
	                }
	            }
	        }

	        for (i = 0; i < MAX_ENEMIES; i++) {
	        	if (enemies[i].active) {
	        		UpdateMovingEntity(&ticks, &enemies[i]);

					if (CheckEntitiesCollisionBB(&player.entity, &enemies[i].entity)) {
				        deadCount = DEATH_TIME;
					}
	        	}
	        }

			if (deadCount > 0) {
				SetAnimationFrame(&player.entity, kDeath, 0, 0);
			}
		} else if (winCount > 0) {
			// Win Map animation.
			CheckPlayerFade();
		}

        if (cpct_isKeyPressed(keys.exit) || currentMap == NUM_MAPS) {
            break;
        }

        UpdateTotalTime();

        // Draw
        cpct_waitVSYNC();

        for (i = 0; i < PLAYER_MAX_BULLETS; i++) {
			CheckMovingEntity(&playerBullets[i]);
        }

        for (i = 0; i < MAX_ENEMIES; i++) {
        	CheckMovingEntity(&enemies[i]);
        	CheckDeathDrawing(&deaths[i]);
        }

        if (!winCount) {
        	DrawEntity(&player.entity);
        }

		DrawTotalTime(0);

        ticks++;
    }
}

void CheckInput(u8 *ticks) {
    u8 leftPressed = 0;
    u8 rightPressed = 0;
    u8 firePressed = 0;
    u8 jumpPressed = 0;

    cpct_scanKeyboard_f();

    if (cpct_isKeyPressed(keys.pause)) {
    	drawPause = 1;
    	pauseTicks = 0;
    	isPaused = 2;
    }

	leftPressed = cpct_isKeyPressed(keys.left);
	rightPressed = cpct_isKeyPressed(keys.right);
    firePressed = cpct_isKeyPressed(keys.fire);
    jumpPressed = cpct_isKeyPressed(keys.jump);

    player.isFiring = firePressed;

	if (jumpPressed == 0) {
		player.jumpBlocked = 0;
	}

	player.entity.prevX = player.entity.x;
	player.entity.prevY = player.entity.y;

	if (rightPressed) {
		if (player.leftSpeed > 0) {
			player.leftSpeed -= 2;
			if (player.leftSpeed < 2 || player.leftSpeed > MAX_SPEED) {
				player.leftSpeed = 0;
			}
		} else if (player.rightSpeed < MAX_SPEED) {
			player.rightSpeed++;
		}
	} else if (leftPressed) {
		if (player.rightSpeed > 0) {
			player.rightSpeed -= 2;
			if (player.rightSpeed < 2 || player.rightSpeed > MAX_SPEED) {
				player.rightSpeed = 0;
			}
		} else if (player.leftSpeed < MAX_SPEED) {
			player.leftSpeed++;
		}
	} else {
		if (player.rightSpeed > 0) {
			if (player.state == GROUNDED) {
				player.rightSpeed -= 2;

				if (player.rightSpeed < 2 || player.rightSpeed > MAX_SPEED) {
					player.rightSpeed = 0;
				}
			} else if ((*ticks & 1) == 0) {
				player.rightSpeed--;
			}
		} else if (player.leftSpeed > 0) {
			if (player.state == GROUNDED) {
				player.leftSpeed -= 2;

				if (player.leftSpeed < 2 || player.leftSpeed > MAX_SPEED) {
					player.leftSpeed = 0;
				}
			} else if ((*ticks & 1) == 0) {
				player.leftSpeed--;
			}
		}
	}

	if (player.state == GROUNDED && jumpPressed) {
		if (player.jumpBlocked == 0) {
			player.jumpBlocked = 1;
			player.state = JUMPING;
			cpct_akp_SFXPlay(1, 13, 48, 0, 0, AY_CHANNEL_B);
			// Parameters: instrument number, volume (0-15), note, speed (0 = original), inverted pitch, channel number (0-2)
		}
	}
}

void UpdatePlayer(u8 *ticks) {
	u8 hasToMove = 1;

	// Horizontal movement.
	if (player.rightSpeed > 0) {
		player.lastSpeedIsRight = 1;
		if (!CheckMapCollision(player.entity.right, player.entity.y, COLLISION_TYPE_PLAYER) && !CheckMapCollision(player.entity.right, player.entity.bottom - 1, COLLISION_TYPE_PLAYER)) {
			if (player.rightSpeed < HALF_MAX_SPEED) {
				if ((*ticks & 1) == 0) {
					hasToMove = 0;
				}
			}

			if (hasToMove) {
				player.entity.x++;
				player.entity.right++;
			}
		} else {
			player.rightSpeed = 0;
		}
	} else if (player.leftSpeed > 0) {
		player.lastSpeedIsRight = 0;
		if (!CheckMapCollision(player.entity.x - 1, player.entity.y, COLLISION_TYPE_PLAYER) && !CheckMapCollision(player.entity.x - 1, player.entity.bottom - 1, COLLISION_TYPE_PLAYER)) {
			if (player.leftSpeed < HALF_MAX_SPEED) {
				if ((*ticks & 1) == 0) {
					hasToMove = 0;
				}
			}

			if (hasToMove) {
				player.entity.x--;
				player.entity.right--;
			}
		} else {
			player.leftSpeed = 0;
		}
	}

	// Vertical movement.
	if (player.state == JUMPING) {
		if (!CheckMapCollision(player.entity.x, player.entity.y - kJumpValues[player.jumpFactor], COLLISION_TYPE_PLAYER) && !CheckMapCollision(player.entity.right - 1, player.entity.y - kJumpValues[player.jumpFactor], COLLISION_TYPE_PLAYER)) {
			player.entity.y -= kJumpValues[player.jumpFactor];
			player.entity.bottom -= kJumpValues[player.jumpFactor];

			if (player.jumpFactor < JUMP_SIZE - 1) {
				player.jumpFactor++;
			} else {
				player.state = FALLING;
				player.jumpFactor = JUMP_SIZE - 1;
			}
		} else {
			player.state = FALLING;
			player.jumpFactor = JUMP_SIZE - 1;

			// Move until it touches.
			while (!CheckMapCollision(player.entity.x, player.entity.y - 1, COLLISION_TYPE_PLAYER) && !CheckMapCollision(player.entity.right - 1, player.entity.y - 1, COLLISION_TYPE_PLAYER)) {
				player.entity.y -= 1;
				player.entity.bottom -= 1;
			}
		}
	} else if (player.state == FALLING) {
		if (!CheckMapCollision(player.entity.x, player.entity.bottom + kJumpValues[player.jumpFactor] - 1, COLLISION_TYPE_PLAYER) && !CheckMapCollision(player.entity.right - 1, player.entity.bottom + kJumpValues[player.jumpFactor] - 1, COLLISION_TYPE_PLAYER)) {
		    player.entity.y += kJumpValues[player.jumpFactor];
			player.entity.bottom += kJumpValues[player.jumpFactor];

			if (player.jumpFactor > 0) {
				player.jumpFactor--;
			}
		} else {
			player.state = GROUNDED;
			player.jumpFactor = 0;

			// Move until it touches.
			while (!CheckMapCollision(player.entity.x, player.entity.bottom, COLLISION_TYPE_PLAYER) && !CheckMapCollision(player.entity.right - 1, player.entity.bottom, COLLISION_TYPE_PLAYER)) {
				player.entity.y += 1;
				player.entity.bottom += 1;
			}
		}
	} else if (player.state == GROUNDED) {
		if (!CheckMapCollision(player.entity.x, player.entity.bottom, COLLISION_TYPE_PLAYER) && !CheckMapCollision(player.entity.right - 1, player.entity.bottom, COLLISION_TYPE_PLAYER)) {
			player.state = FALLING;
			player.jumpFactor = JUMP_SIZE - 5;
		}
	}

    // Fire
    if (player.nextFire > 0) {
        player.nextFire--;
    } else if (player.isFiring) {
        u8 i, collides, fireRight;

        for (i = 0; i < PLAYER_MAX_BULLETS; i++) {
            if (playerBullets[i].active == 0) {

                playerBullets[i].entity.width = HALF_TILE_WIDTH;
                playerBullets[i].entity.height = EIGHTH_TILE_HEIGHT;
				playerBullets[i].entity.animation.frameId = NO_ANIMATION_FRAME_ID;

                if (cpct_isKeyPressed(keys.right)) {
                	fireRight = 1;
                } else if (cpct_isKeyPressed(keys.left)) {
                	fireRight = 0;
                } else if (player.lastSpeedIsRight) {
                	fireRight = 1;
                } else {
                	fireRight = 0;
                }

                if (fireRight) {
                    playerBullets[i].entity.x = player.entity.right;
                    playerBullets[i].entity.y = player.entity.y + HALF_TILE_HEIGHT - 1;
                    playerBullets[i].xSpeed = PLAYER_BULLET_SPEED;
                } else {
                    playerBullets[i].entity.x = player.entity.x - playerBullets[i].entity.width;
                    playerBullets[i].entity.y = player.entity.y + HALF_TILE_HEIGHT - 1;
                    playerBullets[i].xSpeed = -PLAYER_BULLET_SPEED;
                }

                playerBullets[i].entity.right = playerBullets[i].entity.x + playerBullets[i].entity.width;
                playerBullets[i].entity.bottom = playerBullets[i].entity.y + playerBullets[i].entity.height;

                collides = 0;
                if (playerBullets[i].xSpeed > 0) {
                    if ((CheckMapCollisionLaser(playerBullets[i].entity.right, playerBullets[i].entity.y)
                        || CheckMapCollisionLaser(playerBullets[i].entity.right, playerBullets[i].entity.bottom - 1))) {
                        collides = 1;
                    }
                } else if (playerBullets[i].xSpeed < 0) {
                    if (CheckMapCollisionLaser(playerBullets[i].entity.x - 1, playerBullets[i].entity.y)
                        || CheckMapCollisionLaser(playerBullets[i].entity.x - 1, playerBullets[i].entity.bottom - 1)) {
                        collides = 1;
                    }
                }
                player.nextFire = PLAYER_FIRE_RATE;

                cpct_akp_SFXPlay(9, 13, 38, 0, 0, AY_CHANNEL_B);

                if (collides == 0) {
                    playerBullets[i].ySpeed = 0;
                    playerBullets[i].type = 0;	// So does not check bounce.
                    playerBullets[i].erase = 0;

                    playerBullets[i].entity.prevX = playerBullets[i].entity.x;
                    playerBullets[i].entity.prevY = playerBullets[i].entity.y;

                    playerBullets[i].active = 1;
                    break;
                }
            }

        }

    }

	// Animations.
	if (player.state == GROUNDED) {
		if (player.rightSpeed > 0) {
			SetAnimationFrame(&player.entity, kPlayerMoveRight, 1, 0);
		} else if (player.leftSpeed > 0) {
			SetAnimationFrame(&player.entity, kPlayerMoveLeft, 1, 0);
		} else {
			if (player.lastSpeedIsRight) {
				SetAnimationFrame(&player.entity, kPlayerIdleRight, 1, 0);
			} else {
				SetAnimationFrame(&player.entity, kPlayerIdleLeft, 1, 0);
			}
		}
	} else if (player.state == JUMPING) {
		if (player.rightSpeed > 0) {
			SetAnimationFrame(&player.entity, kPlayerJumpRight, 1, 0);
		} else if (player.leftSpeed > 0) {
			SetAnimationFrame(&player.entity, kPlayerJumpLeft, 1, 0);
		} else {
			if (player.lastSpeedIsRight) {
				SetAnimationFrame(&player.entity, kPlayerJumpRight, 1, 0);
			} else {
				SetAnimationFrame(&player.entity, kPlayerJumpLeft, 1, 0);
			}
		}
	} else if (player.state == FALLING) {
		if (player.rightSpeed > 0) {
			SetAnimationFrame(&player.entity, kPlayerFallRight, 1, 0);
		} else if (player.leftSpeed > 0) {
			SetAnimationFrame(&player.entity, kPlayerFallLeft, 1, 0);
		} else {
			if (player.lastSpeedIsRight) {
				SetAnimationFrame(&player.entity, kPlayerFallRight, 1, 0);
			} else {
				SetAnimationFrame(&player.entity, kPlayerFallLeft, 1, 0);
			}
		}
	}
}

void ResetPlayer(void) {
	player.entity.x = player.startX;
	player.entity.y = player.startY;
	player.entity.prevX = player.entity.x;
	player.entity.prevY = player.entity.y;
    player.entity.width = TILE_WIDTH;
    player.entity.height = TILE_HEIGHT;
	player.entity.right = player.entity.x + player.entity.width;
	player.entity.bottom = player.entity.y + player.entity.height;

	player.state = FALLING;
	player.jumpFactor = JUMP_SIZE - 1;
	player.jumpBlocked = 0;
	player.leftSpeed = 0;
	player.rightSpeed = 0;
}

void ChangeMap() {
	currentMap++;

	if (currentMap < NUM_MAPS) {
		LoadMap(currentMap);
	}
}

void LoadMap(u8 mapNumber) {
    u8 i, j, leftPixel, rightPixel, n, e, s, w, enemyIndex;

	cpct_clearScreen_f64(0);

    // Each byte contains two tiles.
    for (i = 0; i < MAP_HEIGHT; i++) {
        for (j = 0; j < MAP_HALF_WIDTH; j++) {
            leftPixel = GetLeftPixelValue((u8)kMaps[mapNumber][i][j]);
            rightPixel = GetRightPixelValue((u8)kMaps[mapNumber][i][j]);

            map[j * 2][i] = leftPixel;
            map[j * 2 + 1][i] = rightPixel;
        }
    }

    // Reset enemies and bullets, so those from previous levels do not appear in the new.
    enemyIndex = 0;

    for (i = 0; i < MAX_ENEMIES; i++) {
    	enemies[i].active = 0;
    	enemies[i].spawns = 0;
    }

    for (i = 0; i < PLAYER_MAX_BULLETS; i++) {
    	playerBullets[i].active = 0;
    }

	numBlocks = 0;

    // Draw tiles.
    for (i = 0; i < MAP_WIDTH; i++) {
        for (j = 0; j < MAP_HEIGHT; j++) {
            if (map[i][j] > 0) {
                if (map[i][j] == MAP_PLAYER_START) {
					player.startX = i * TILE_WIDTH;
					player.startY = j * TILE_HEIGHT;
					map[i][j] = MAP_VOID;	// No collide.
				} else if (map[i][j] == MAP_CHASER) {
					if (enemyIndex < MAX_ENEMIES) {
						SpawnEnemy(&enemies[enemyIndex++], i * TILE_WIDTH, j * TILE_HEIGHT, ENEMY_TYPE_CHASES | ENEMY_TYPE_MAP_INMUNE | ENEMY_TYPE_RESPAWNS);
					}
					map[i][j] = MAP_VOID;	// No collide.
				} else if (map[i][j] == MAP_DEFORM_BLOCK || map[i][j] == MAP_BREAKABLE_BLOCK) {
					if (numBlocks < MAX_BLOCKS) {
						blocks[numBlocks].x = i;
						blocks[numBlocks].y = j;
						blocks[numBlocks].time = 0;
						blocks[numBlocks].type = map[i][j];
						numBlocks++;
					}
				} else if (map[i][j] == MAP_INMUNE_CHASER) {
					if (enemyIndex < MAX_ENEMIES) {
						SpawnEnemy(&enemies[enemyIndex++], i * TILE_WIDTH, j * TILE_HEIGHT, ENEMY_TYPE_CHASES | ENEMY_TYPE_MAP_INMUNE | ENEMY_TYPE_LASER_INMUNE);
					}
					map[i][j] = MAP_VOID;	// No collide.
				} else if (map[i][j] == MAP_BOUNCER_X) {
					if (enemyIndex < MAX_ENEMIES) {
						SpawnEnemy(&enemies[enemyIndex++], i * TILE_WIDTH, j * TILE_HEIGHT, ENEMY_TYPE_X | ENEMY_TYPE_BOUNCES | ENEMY_TYPE_MAP_INMUNE | ENEMY_TYPE_LASER_INMUNE);
					}
					map[i][j] = MAP_VOID;	// No collide.
				} else if (map[i][j] == MAP_BOUNCER_Y) {
					if (enemyIndex < MAX_ENEMIES) {
						SpawnEnemy(&enemies[enemyIndex++], i * TILE_WIDTH, j * TILE_HEIGHT, ENEMY_TYPE_Y | ENEMY_TYPE_BOUNCES | ENEMY_TYPE_MAP_INMUNE | ENEMY_TYPE_LASER_INMUNE);
					}
					map[i][j] = MAP_VOID;	// No collide.
				} else if (map[i][j] == MAP_BOUNCER_XY) {
					if (enemyIndex < MAX_ENEMIES) {
						SpawnEnemy(&enemies[enemyIndex++], i * TILE_WIDTH, j * TILE_HEIGHT, ENEMY_TYPE_X | ENEMY_TYPE_Y | ENEMY_TYPE_BOUNCES | ENEMY_TYPE_MAP_INMUNE | ENEMY_TYPE_LASER_INMUNE);
					}
					map[i][j] = MAP_VOID;	// No collide.
				}
            }
        }
    }

    // Draw basic tiles.
    for (i = 0; i < MAP_WIDTH; i++) {
        for (j = 0; j < MAP_HEIGHT; j++) {
            if (map[i][j] > 0) {
            	n = 0;
            	e = 0;
            	s = 0;
            	w = 0;

            	// IMPROVE If deform and breakable are 1 and 2, the check would be map[][] > 2
                if (j > 0) {
                	if (map[i][j - 1] > 0 && map[i][j - 1] != MAP_DEFORM_BLOCK && map[i][j - 1] != MAP_BREAKABLE_BLOCK) {
                		n = 1;
                	}
                } else {
                	n = 1;
                }
                
                if (j < MAP_HEIGHT - 1) {
                	if (map[i][j + 1] > 0 && map[i][j + 1] != MAP_DEFORM_BLOCK && map[i][j + 1] != MAP_BREAKABLE_BLOCK) {
                		s = 1;
                	}
                } else {
                	s = 1;
                }

                if (i > 0) {
                	if (map[i - 1][j] > 0 && map[i - 1][j] != MAP_DEFORM_BLOCK && map[i - 1][j] != MAP_BREAKABLE_BLOCK) {
                		w = 1;
                	}
                } else {
                	w = 1;
                }
                
                if (i < MAP_WIDTH - 1) {
                	if (map[i + 1][j] > 0 && map[i + 1][j] != MAP_DEFORM_BLOCK && map[i + 1][j] != MAP_BREAKABLE_BLOCK) {
                		e = 1;
                	}
                } else {
                	e = 1;
                }

                // Check the type of tile.
                if (map[i][j] == MAP_BASIC) {
	                if (!s) {
		                if (w && n && e) {
		                	DrawTile(kTileBasicS, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);
		                } else if (w && n) {
		            		DrawTile(kTileBasicES, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);
		                } else if (w) {
		            		DrawTile(kTileBasicNES, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);
		                }
	                }

	                if (!w) {
	                	if (n && e && s) {
	                		DrawTile(kTileBasicW, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);
	                	} else if (n && e) {
	                		DrawTile(kTileBasicSW, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);
	                	} else if (n) {
	                		DrawTile(kTileBasicESW, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);
	                	}
	                }

	                if (!n) {
	                	if (e && s && w) {
	                		DrawTile(kTileBasicN, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);
	                	} else if (e && s) {
	                		DrawTile(kTileBasicWN, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);
	                	} else if (e && w) {
	                		DrawTile(kTileBasicNS, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);
	                	} else if (e) {
	                		DrawTile(kTileBasicSWN, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);
	                	}
	                }

	                if (!e) {
	                	if (s && w && n) {
	                		DrawTile(kTileBasicE, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);
	                	} else if (s && w) {
	                		DrawTile(kTileBasicNE, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);
	                	} else if (s && n) {
	                		DrawTile(kTileBasicEW, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);
	                	} else if (s) {
	                		DrawTile(kTileBasicWNE, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);
	                	}
	                }

	                if (!n && !e && !s && !w) {
	                	DrawTile(kTileBasicNESW, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);
	                } else if (n && e && s && w) {
	                	DrawTile(kTileBasicVoid, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);
	                }
                } else if (map[i][j] == MAP_LASER) {
                	if (!n) {
                		DrawTile(kTileLaserN, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);
                	}

                	if (!e) {
                		DrawTile(kTileLaserE, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);
                	}

                	if (!s) {
                		DrawTile(kTileLaserS, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);
                	}

                	if (!w) {
                		DrawTile(kTileLaserW, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);
                	}
                } else if (map[i][j] == MAP_TURRET) {
                	if (!n) {
                		DrawTile(kTurretN, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);

						if (enemyIndex < MAX_ENEMIES) {
							SpawnEnemy(&enemies[enemyIndex++], i * TILE_WIDTH + HALF_TILE_WIDTH, j * TILE_HEIGHT - HALF_TILE_HEIGHT, ENEMY_TYPE_Y | ENEMY_TYPE_RESPAWNS | ENEMY_TYPE_INVERT);
						}
                	}

                	if (!e) {
                		DrawTile(kTurretE, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);

						if (enemyIndex < MAX_ENEMIES) {
							SpawnEnemy(&enemies[enemyIndex++], i * TILE_WIDTH + TILE_WIDTH, j * TILE_HEIGHT + HALF_TILE_HEIGHT - 2, ENEMY_TYPE_X | ENEMY_TYPE_RESPAWNS);
						}
                	}

                	if (!s) {
                		DrawTile(kTurretS, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);

						if (enemyIndex < MAX_ENEMIES) {
							SpawnEnemy(&enemies[enemyIndex++], i * TILE_WIDTH + HALF_TILE_WIDTH, j * TILE_HEIGHT + TILE_HEIGHT, ENEMY_TYPE_Y | ENEMY_TYPE_RESPAWNS);
						}
                	}

                	if (!w) {
                		DrawTile(kTurretW, CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);

						if (enemyIndex < MAX_ENEMIES) {
							SpawnEnemy(&enemies[enemyIndex++], i * TILE_WIDTH - HALF_TILE_WIDTH, j * TILE_HEIGHT + HALF_TILE_HEIGHT - 2, ENEMY_TYPE_X | ENEMY_TYPE_RESPAWNS | ENEMY_TYPE_INVERT);
						}
                	}
                } else {
                	DrawTile(GetTile(map[i][j]), CPCT_VMEM_START + i * 4 + SCR_WIDTH * j * 2);
                }
            }
        }
    }

	ResetDeaths();
    ResetPlayer();

    // HUD
    DrawCurrentMap();
	DrawTotalTime(1);
	DrawTotalDefeats();
	DrawText("DEFEATS ", 15, 192);
	DrawText("TIME ", 57, 192);
	DrawCode();

	cpct_waitVSYNC();
}

u8* GetTile(u8 i) {
	switch (i) {
		case MAP_END: 				return kTileEnd;
		case MAP_DEFORM_BLOCK: 		return kTileDeformBlock01;
		case MAP_BREAKABLE_BLOCK: 	return kTileBreakableBlock01;
		default: 					return kTileBasicVoid;
	}
}

u8 CheckMapCollision(u8 x, u8 y, u8 type) {
	x = x >> 2;
	y = y >> 4;

	// Bounds.
	if (x == 0xFF) {
		return 1;
	}

	if (y == 0xFF) {
		return 1;
	}

	if (x >= MAP_WIDTH) {
		return 1;
	}

	if (y >= MAP_HEIGHT) {
		return 1;
	}

	if (type == COLLISION_TYPE_PLAYER) {
		if (map[x][y] == MAP_LASER) {
			deadCount = DEATH_TIME;	// Wait and reset.
		} else if (map[x][y] == MAP_END) {
			winCount = WIN_TIME;	// Wait and reset.
		} else if (map[x][y] == MAP_DEFORM_BLOCK) {
			ActivateBlock(x, y);
		}
	} else if (type == COLLISION_TYPE_LASER) {
		if (map[x][y] == MAP_BREAKABLE_BLOCK) {
			ActivateBlock(x, y);
		}
	}

	return map[x][y];
}

u8 CheckMapCollisionBasic(u8 x, u8 y) {
	x = x >> 2;
	y = y >> 4;

	return map[x][y];
}

// This is used to activate blocks when the laser does not appear (player too close to the block).
u8 CheckMapCollisionLaser(u8 x, u8 y) {
	x = x >> 2;
	y = y >> 4;

	if (map[x][y] == MAP_BREAKABLE_BLOCK) {
		ActivateBlock(x, y);
	}

	return map[x][y];
}

void UpdateMovingEntity(u8 *ticks, TMovingEntity *movingEntity) {
	u8 collisionType;

	movingEntity->entity.prevX = movingEntity->entity.x;
	movingEntity->entity.prevY = movingEntity->entity.y;

	if ((*ticks & 1) == 0) {
		// Chases, moves slowly towards the player.
		if (movingEntity->type & ENEMY_TYPE_CHASES) {
			if ((*ticks & 3) == 0) {
				if (player.entity.x > movingEntity->entity.x) {
					movingEntity->xSpeed = 1;
				} else if ((player.entity.x < movingEntity->entity.x)) {
					movingEntity->xSpeed = -1;
				} else {
					movingEntity->xSpeed = 0;
				}
			} else {
				movingEntity->xSpeed = 0;
			}

			if (player.entity.y > movingEntity->entity.y) {
				movingEntity->ySpeed = 2;
			} else if (player.entity.y < movingEntity->entity.y) {
				movingEntity->ySpeed = -2;
			} else {
				movingEntity->ySpeed = 0;
			}
		}

		// Check Collision Type. EIGHTH_TILE_HEIGHT is the height of the player laser.
		if (movingEntity->entity.height == EIGHTH_TILE_HEIGHT) {
			collisionType = COLLISION_TYPE_LASER;
		} else {
			collisionType = COLLISION_TYPE_ENEMY;
		}

		// Horizontal movement.
		if (movingEntity->xSpeed > 0) {
			if (!CheckMapCollisionBasic(movingEntity->entity.right + movingEntity->xSpeed, movingEntity->entity.y)
				&& !CheckMapCollisionBasic(movingEntity->entity.right + movingEntity->xSpeed, movingEntity->entity.bottom - 1)) {
				movingEntity->entity.x += movingEntity->xSpeed;
				movingEntity->entity.right += movingEntity->xSpeed;
			} else {
                // Move until it touches.
                while (!CheckMapCollision(movingEntity->entity.right, movingEntity->entity.y, collisionType)
                	    && !CheckMapCollision(movingEntity->entity.right, movingEntity->entity.bottom - 1, collisionType)) {
    				movingEntity->entity.x += 1;
    				movingEntity->entity.right += 1;
    			}

                if (movingEntity->type & ENEMY_TYPE_BOUNCES) {
                    movingEntity->xSpeed *= -1;
                } else if (!(movingEntity->type & ENEMY_TYPE_MAP_INMUNE)) {
                    movingEntity->xSpeed = 0;
                    movingEntity->erase = 2;
                }
			}
		} else if (movingEntity->xSpeed < 0) {
			if (!CheckMapCollisionBasic(movingEntity->entity.x - 1 + movingEntity->xSpeed, movingEntity->entity.y)
				&& !CheckMapCollisionBasic(movingEntity->entity.x - 1  + movingEntity->xSpeed, movingEntity->entity.bottom - 1)) {
				movingEntity->entity.x += movingEntity->xSpeed;
				movingEntity->entity.right += movingEntity->xSpeed;
			} else {
                // Move until it touches.
                while (!CheckMapCollision(movingEntity->entity.x - 1, movingEntity->entity.y, collisionType)
                	   && !CheckMapCollision(movingEntity->entity.x - 1, movingEntity->entity.bottom - 1, collisionType)) {
    				movingEntity->entity.x -= 1;
    				movingEntity->entity.right -= 1;
    			}

                if (movingEntity->type & ENEMY_TYPE_BOUNCES) {
                    movingEntity->xSpeed *= -1;
                } else if (!(movingEntity->type & ENEMY_TYPE_MAP_INMUNE)) {
                    movingEntity->xSpeed = 0;
                    movingEntity->erase = 2;
                }
			}
		}

		// Vertical movement.
		if (movingEntity->ySpeed < 0) {
			if (!CheckMapCollisionBasic(movingEntity->entity.x, movingEntity->entity.y + movingEntity->ySpeed)
				&& !CheckMapCollisionBasic(movingEntity->entity.right - 1, movingEntity->entity.y + movingEntity->ySpeed)) {
				movingEntity->entity.y += movingEntity->ySpeed;
				movingEntity->entity.bottom += movingEntity->ySpeed;
			} else {
                while (!CheckMapCollision(movingEntity->entity.x, movingEntity->entity.y - 1, collisionType)
                	   && !CheckMapCollision(movingEntity->entity.right - 1, movingEntity->entity.y - 1, collisionType)) {
    				movingEntity->entity.y -= 1;
    				movingEntity->entity.bottom -= 1;
    			}

                if (movingEntity->type & ENEMY_TYPE_BOUNCES) {
                    movingEntity->ySpeed *= -1;
                } else if (!(movingEntity->type & ENEMY_TYPE_MAP_INMUNE)) {
	    			movingEntity->ySpeed = 0;
	                movingEntity->erase = 2;
                }
			}
		} else if (movingEntity->ySpeed > 0) {
			if (!CheckMapCollisionBasic(movingEntity->entity.x, movingEntity->entity.bottom + movingEntity->ySpeed - 1)
				&& !CheckMapCollisionBasic(movingEntity->entity.right - 1, movingEntity->entity.bottom + movingEntity->ySpeed - 1)) {
				movingEntity->entity.y += movingEntity->ySpeed;
				movingEntity->entity.bottom += movingEntity->ySpeed;
			} else {
                while (!CheckMapCollision(movingEntity->entity.x, movingEntity->entity.bottom, collisionType)
                	   && !CheckMapCollision(movingEntity->entity.right - 1, movingEntity->entity.bottom, collisionType)) {
    				movingEntity->entity.y += 1;
    				movingEntity->entity.bottom += 1;
    			}

                if (movingEntity->type & ENEMY_TYPE_BOUNCES) {
                    movingEntity->ySpeed *= -1;
                } else if (!(movingEntity->type & ENEMY_TYPE_MAP_INMUNE)) {
	    			movingEntity->ySpeed = 0;
	                movingEntity->erase = 2;
                }
			}
		}
	}
}

void CheckMovingEntity(TMovingEntity *movingEntity) {
    if (movingEntity->erase) {
    	movingEntity->erase--;

    	if (movingEntity->erase == 0) {
            EraseEntity(&movingEntity->entity);
			movingEntity->active = 0;

            if (movingEntity->type & ENEMY_TYPE_RESPAWNS) {
            	if (movingEntity->type &ENEMY_TYPE_CHASES) {
					SpawnDeath(movingEntity->entity.x, movingEntity->entity.y);
            	}

				SpawnEnemy(movingEntity, movingEntity->startX, movingEntity->startY, movingEntity->type);
            } else {
            	// Eight Tile Height is the player laser.
            	if (movingEntity->entity.height != EIGHTH_TILE_HEIGHT) {
            		SpawnDeath(movingEntity->entity.x, movingEntity->entity.y);
            	}
            }
    	} else {
    		DrawEntity(&movingEntity->entity);
    	}

    } else if (movingEntity->active) {
        DrawEntity(&movingEntity->entity);
    }
}

void ResetMovingEntities() {
	u8 i;

    for (i = 0; i < MAX_ENEMIES; i++) {
    	if (enemies[i].spawns) {
    		EraseEntity(&enemies[i].entity);
    		SpawnEnemy(&enemies[i], enemies[i].startX, enemies[i].startY, enemies[i].type);
    	}
    }
}

void SpawnEnemy(TMovingEntity *movingEntity, u8 x, u8 y, u8 type) {
	movingEntity->active = 1;
	movingEntity->spawns = 1;
	movingEntity->startX = x;
	movingEntity->startY = y;
	movingEntity->entity.x = movingEntity->startX;
	movingEntity->entity.y = movingEntity->startY;
	movingEntity->entity.prevX = movingEntity->entity.x;
	movingEntity->entity.prevY = movingEntity->entity.y;

	// Dimensions and animations of the enemy.
	if ((type & ENEMY_TYPE_X) && (type & ENEMY_TYPE_RESPAWNS)) {
		movingEntity->entity.width = HALF_TILE_WIDTH;
		movingEntity->entity.height = QUARTER_TILE_HEIGHT;
		movingEntity->entity.animation.frameId = NO_ANIMATION_FRAME_ID;
	} else if ((type & ENEMY_TYPE_Y) && (type & ENEMY_TYPE_RESPAWNS)) {
		movingEntity->entity.width = QUARTER_TILE_WIDTH;
		movingEntity->entity.height = HALF_TILE_HEIGHT;
		movingEntity->entity.animation.frameId = NO_ANIMATION_FRAME_ID;
	} else {
		movingEntity->entity.width = TILE_WIDTH;
		movingEntity->entity.height = TILE_HEIGHT;

		if (type & ENEMY_TYPE_BOUNCES) {
			if (currentMap == 0) {
				// Phantis asteroid if the current map is the easter egg.
				SetAnimationFrame(&movingEntity->entity, kEnemyAsteroid, 1, 1);
			} else {
				SetAnimationFrame(&movingEntity->entity, kEnemyBouncer, 1, 1);	
			}
		} else if (type & ENEMY_TYPE_LASER_INMUNE) {
			SetAnimationFrame(&movingEntity->entity, kEnemyChaserInmune, 1, 1);
		} else {
			SetAnimationFrame(&movingEntity->entity, kEnemyChaser, 1, 1);
		}
	}

	movingEntity->entity.right = movingEntity->entity.x + movingEntity->entity.width;
	movingEntity->entity.bottom = movingEntity->entity.y + movingEntity->entity.height;
	movingEntity->type = type;

	movingEntity->xSpeed = 0;
	movingEntity->ySpeed = 0;

	if (type & ENEMY_TYPE_X) {
		if (type & ENEMY_TYPE_RESPAWNS) {
			movingEntity->xSpeed = 2;
		} else {
			movingEntity->xSpeed = 1;
		}

		if (type & ENEMY_TYPE_INVERT) {
			movingEntity->xSpeed *= -1;					
		}
	}

	if (type & ENEMY_TYPE_Y) {
		if (type & ENEMY_TYPE_RESPAWNS) {
			movingEntity->ySpeed = 8;
		} else {
			movingEntity->ySpeed = 4;
		}

		if (type & ENEMY_TYPE_INVERT) {
			movingEntity->ySpeed *= -1;					
		}
	}
}

void SpawnDeath(u8 x, u8 y) {
	u8 i;

	for (i = 0; i < MAX_ENEMIES; i++) {
		if (deaths[i].animation.frameId == NO_ANIMATION_FRAME_ID) {
			cpct_akp_SFXPlay(2, 13, 44, 0, 0, AY_CHANNEL_B);
			deaths[i].x = x;
			deaths[i].y = y;
			deaths[i].prevX = deaths[i].x;
			deaths[i].prevY = deaths[i].y;
		    deaths[i].width = TILE_WIDTH;
		    deaths[i].height = TILE_HEIGHT;
			deaths[i].right = deaths[i].x + deaths[i].width;
			deaths[i].bottom = deaths[i].y + deaths[i].height;
			SetAnimationFrame(&deaths[i], kDeath, 0, 1);
			break;
		}
	}
}

void CheckDeathDrawing(TEntity *entity) {
	if (entity->animation.frameId != NO_ANIMATION_FRAME_ID) {
		if (entity->animation.frameId == 4 && entity->animation.time == 1) {	// This indicates that the animation has finished.
			entity->animation.frameId = NO_ANIMATION_FRAME_ID;
		} else {
			DrawEntity(entity);
		}
	}
}

void ResetDeaths() {
	for (u8 i = 0; i < MAX_ENEMIES; i++) {
		if (deaths[i].animation.frameId != NO_ANIMATION_FRAME_ID) {
			EraseEntity(&deaths[i]);
			deaths[i].animation.frameId = NO_ANIMATION_FRAME_ID;
		}
	}
}

void ActivateBlock(u8 x, u8 y) {
	u8 i = 0;

	for (i = 0; i < numBlocks; i++) {
		if (blocks[i].x == x && blocks[i].y == y) {
			if (blocks[i].time == 0) {
				cpct_akp_SFXPlay(4, 13, 46, 0, 0, AY_CHANNEL_B);
				blocks[i].time = BLOCK_TIME;

				if (blocks[i].type == MAP_DEFORM_BLOCK) {
					DrawTile(kTileDeformBlock02, CPCT_VMEM_START + x * 4 + SCR_WIDTH * y * 2);
				} else {
					DrawTile(kTileBreakableBlock02, CPCT_VMEM_START + x * 4 + SCR_WIDTH * y * 2);
				}


			}
		}
	}
}

void UpdateBlocks() {
	u8 i = 0;

	for (i = 0; i < numBlocks; i++) {
		if (blocks[i].time > 0) {
			blocks[i].time--;

			if (blocks[i].time == 0) {
				cpct_akp_SFXPlay(5, 13, 56, 0, 0, AY_CHANNEL_B);
				DrawTile(kTileBasicVoid, CPCT_VMEM_START + blocks[i].x * 4 + SCR_WIDTH * blocks[i].y * 2);
				map[blocks[i].x][blocks[i].y] = MAP_VOID;
			}
		}
	}
}

void ResetBlocks() {
	u8 i = 0;

	for (i = 0; i < numBlocks; i++) {
		DrawTile(GetTile(blocks[i].type), CPCT_VMEM_START + blocks[i].x * 4 + SCR_WIDTH * blocks[i].y * 2);
		map[blocks[i].x][blocks[i].y] = blocks[i].type;
		blocks[i].time = 0;
	}
}

void DrawCurrentMap() {
	u8 x = 1;

	DrawText("MAP ", x, 192);
	x = x + 7;
	DrawComposedNumber(currentMap, x, 192, 0);
}

void UpdateTotalTime() {
	timeCount--;

	if (timeCount == 0) {
		timeCount = 50;
		seconds++;

		if (seconds == 60) {
			seconds = 0;
			minutes++;
		}
	}
}

void DrawTotalTime(u8 forceDraw) {
	if (timeCount == 50 || forceDraw) {
		u8 x = 76;

		if (seconds < 10) {
			DrawComposedNumber(seconds, x, 192, 1);
			x = x - 2;

			DrawComposedNumber(0, x, 192, 1);
		} else {
			DrawComposedNumber(seconds, x, 192, 1);
			x = x - 2;
		}

		x = x - 2;

		DrawText(":", x, 192);

		x = x - 2;
		DrawComposedNumber(minutes, x, 192, 1);
	}
}

void DrawTotalDefeats() {
	// TODO ¿Meter hardcore mode (solo 1 vida)?
	DrawComposedNumber(totalDefeats, 30, 192, 0);
}

void DrawCode() {
	u16 code = 0;

	if (currentMap == 11) {
		code = CODE_CHECKPOINT_1;
	} else if (currentMap == 21) {
		code = CODE_CHECKPOINT_2;
	} else if (currentMap == 31) {
		code = CODE_CHECKPOINT_3;
	} else if (currentMap == 41) {
		code = CODE_CHECKPOINT_4;
	}

	if (code > 0) {
		DrawText("CODE ", 31, 0);
		DrawComposedNumber(code, 40, 0, 0);
	}
}

u8 CheckDeath() {
	if (deadCount > 0) {
		if (deadCount == DEATH_TIME) {
			cpct_akp_SFXPlay(2, 13, 44, 0, 0, AY_CHANNEL_B);
			totalDefeats++;
		}

		deadCount--;
		winCount = 0;	// Death is stronger than winning.

		if (deadCount == 0) {
			ResetPlayer();
			ResetMovingEntities();
			ResetBlocks();
			DrawTotalDefeats();
			ResetDeaths();
		}

		return 1;
	} else {
		return 0;
	}
}

u8 CheckWin() {
	if (deadCount > 0) {	// Death is stronger than winning.
		return 0;
	}

	if (winCount > 0) {
		if (winCount == WIN_TIME) {
			cpct_akp_SFXPlay(3, 13, 50, 4, 0, AY_CHANNEL_B);
		}

		winCount--;

		if (winCount == 0) {
			ChangeMap();
		}

		return 1;
	} else {
		return 0;
	}
}

void CheckPlayerFade() {
	u8 *pscr;

	if (winCount == 45) {
		pscr = cpct_getScreenPtr(CPCT_VMEM_START, player.entity.x, player.entity.y + EIGHTH_TILE_HEIGHT);
		cpct_drawSolidBox(pscr, cpct_px2byteM0(COLOR_BLACK, COLOR_BLACK), player.entity.width, EIGHTH_TILE_HEIGHT);
	} else if (winCount == 40) {
		pscr = cpct_getScreenPtr(CPCT_VMEM_START, player.entity.x, player.entity.y + QUARTER_TILE_HEIGHT + EIGHTH_TILE_HEIGHT);
		cpct_drawSolidBox(pscr, cpct_px2byteM0(COLOR_BLACK, COLOR_BLACK), player.entity.width, EIGHTH_TILE_HEIGHT);
	} else if (winCount == 35) {
		pscr = cpct_getScreenPtr(CPCT_VMEM_START, player.entity.x, player.entity.y + HALF_TILE_HEIGHT + EIGHTH_TILE_HEIGHT);
		cpct_drawSolidBox(pscr, cpct_px2byteM0(COLOR_BLACK, COLOR_BLACK), player.entity.width, EIGHTH_TILE_HEIGHT);
	} else if (winCount == 30) {
		pscr = cpct_getScreenPtr(CPCT_VMEM_START, player.entity.x, player.entity.y + HALF_TILE_HEIGHT + QUARTER_TILE_HEIGHT + EIGHTH_TILE_HEIGHT);
		cpct_drawSolidBox(pscr, cpct_px2byteM0(COLOR_BLACK, COLOR_BLACK), player.entity.width, EIGHTH_TILE_HEIGHT);
	} else 	if (winCount == 25) {
		pscr = cpct_getScreenPtr(CPCT_VMEM_START, player.entity.x, player.entity.y);
		cpct_drawSolidBox(pscr, cpct_px2byteM0(COLOR_BLACK, COLOR_BLACK), player.entity.width, EIGHTH_TILE_HEIGHT);
	} else if (winCount == 20) {
		pscr = cpct_getScreenPtr(CPCT_VMEM_START, player.entity.x, player.entity.y + QUARTER_TILE_HEIGHT);
		cpct_drawSolidBox(pscr, cpct_px2byteM0(COLOR_BLACK, COLOR_BLACK), player.entity.width, EIGHTH_TILE_HEIGHT);
	} else if (winCount == 15) {
		pscr = cpct_getScreenPtr(CPCT_VMEM_START, player.entity.x, player.entity.y + HALF_TILE_HEIGHT);
		cpct_drawSolidBox(pscr, cpct_px2byteM0(COLOR_BLACK, COLOR_BLACK), player.entity.width, EIGHTH_TILE_HEIGHT);
	} else if (winCount == 10) {
		pscr = cpct_getScreenPtr(CPCT_VMEM_START, player.entity.x, player.entity.y + HALF_TILE_HEIGHT + QUARTER_TILE_HEIGHT);
		cpct_drawSolidBox(pscr, cpct_px2byteM0(COLOR_BLACK, COLOR_BLACK), player.entity.width, EIGHTH_TILE_HEIGHT);
	}
}

void CheckPause() {
	u8 *pscr;

	if (pauseTicks == 0) {
		if (drawPause) {
			DrawText("PAUSED", 39, 192);
			drawPause = 0;
		} else {
			pscr = cpct_getScreenPtr(CPCT_VMEM_START, 39, 192);
			cpct_drawSolidBox(pscr, cpct_px2byteM0(COLOR_BLACK, COLOR_BLACK), 12, 8);
			drawPause = 1;
		}

		pauseTicks = 64;
	} else {
		pauseTicks--;
	}

	cpct_scanKeyboard_f();

	if (cpct_isKeyPressed(keys.pause)) {
		WaitKeyUp(keys.pause);
		isPaused--;

		if (isPaused == 0) {
			pscr = cpct_getScreenPtr(CPCT_VMEM_START, 39, 192);
			cpct_drawSolidBox(pscr, cpct_px2byteM0(COLOR_BLACK, COLOR_BLACK), 12, 8);
		}
	}

	cpct_waitVSYNC();
}